
Object.defineProperty(exports, '__esModule', { value: true });
const types_1 = require('../types/types');
const helper_1 = require('./helper');

let groupNumber = 0;
let Location;
(function(Location) {
	Location[(Location.internal = 0)] = 'internal';
	Location[(Location.exports = 1)] = 'exports';
	Location[(Location.imports = 2)] = 'imports';
	Location[(Location.importArr = 3)] = 'importArr';
	Location[(Location.ExportArr = 4)] = 'ExportArr';
})(Location || (Location = {}));
function getGroupID(name) {
	const nme = name || 'none';
	return nme + (++groupNumber).toString();
}
function getName(code, indx) {
	const result = {
		name: '',
		full: '',
		pre: '',
	};
	let i = indx;
	let space = '';
	const len = code.length;
	while (i < len && /\{|\(|=|;|:/.test(code[i]) === false) {
		if (/\s/.test(code[i])) space += code[i];
		else {
			if (space) {
				result.pre += result.name + space;
				result.full += space;
				result.name = '';
				space = '';
			}
			result.name += code[i];
			result.full += code[i];
		}
		i++;
	}
	if (code[i] === '(') result.codeType = types_1.codeType.roundBracket;
	if (code[i] === '{') result.codeType = types_1.codeType.curlyBracket;
	if (code[i] === '=') result.codeType = types_1.codeType.equal;
	if (result.codeType != null) result.full += `${space}${code[i]}`;
	return result;
}
function getExportValues(modGroup) {
	const result = modGroup;
	const keys = Object.keys(modGroup);
	let counter = 0;
	const getExport = (name, from) => {
		counter++;
		const mod = modGroup[from];
		if (!mod || counter > 15) return false;
		if (mod.internals[name]) return mod.internals[name];
		if (mod.exports[name]) {
			const nme = mod.exports[name].name;
			if (!mod.exports[name].from) {
				if (nme !== name && mod.internals[nme]) return mod.internals[nme];
				if (mod.imports[nme]) {
					if (mod.imports[nme].value != null) return mod.imports[nme].value;
					result[from].imports[nme].value = getExport(
						mod.imports[nme].name,
						mod.imports[nme].from,
					);
					return result[from].imports[nme].value;
				}
				let arr = mod.exportStr || [];
				if (mod.importStr) arr.push(...mod.importStr);
				arr = arr.filter((val, indx, array) => array.indexOf(val) === indx);
				for (let i = 0, len = arr.length; i < len; i++) {
					const res = getExport(nme, arr[i]);
					if (res) return res;
				}
			} else return getExport(nme, mod.exports[name].from);
		}
		return false;
	};
	keys.forEach(key => {
		const mod = modGroup[key];
		const exportKeys = Object.keys(mod.exports);
		exportKeys.forEach(exprt => {
			counter = 0;
			const ex = mod.exports[exprt];
			if (!ex.all) result[key].exports[exprt].value = getExport(exprt, key);
		});
	});
	return result;
}
function rename(modGroup, options) {
	const result = modGroup;
	const keys = Object.keys(modGroup);
	if (options.all != null) return result;
	const amend = val => {
		if (val.owner) result[val.owner].used = true;
		if (result[val.owner].internals[val.name])
			result[val.owner].internals[val.name].used = true;
	};
	const alter = (code, key) => {
		let c = code;
		const replaceReg = /\n--replace(.*?)=(.*)?\[(all|grp)?(ex|im)\]/;
		const exKeys = result[key].exports ? Object.keys(result[key].exports) : [];
		while (replaceReg.test(c)) {
			const [, id, name, t] = replaceReg.exec(c);
			c = c.replace(replaceReg, '');
			if (t === 'all') {
				if (name) {
					if (result[name]) result[name].allused = true;
					if (!result[key].includeArr) result[key].includeArr = [name];
					else result[key].includeArr.push(name);
				}
				continue;
			}
			exKeys.forEach(exkey => {
				const ex = result[key].exports[exkey];
				if (ex.positionID === id) {
					if (ex.value) amend(ex.value);
					else if (ex.all && ex.from && result[ex.from]) result[ex.from].allused = true;
					result[key].exports[exkey].include = true;
				}
			});
		}
	};
	keys.forEach(key => {
		const mod = modGroup[key];
		const modReg = /(\r?\n?\r?[ \t]*)?\/\*+\s*\$mod=?\s*(['"])?(.*?)['"]?\s*\*\/\s*?(((.|\s)*?)(\/\*+\s*\$modend\s*\*\/)|((.|\s)*))?/;
		let modCode = mod.code;
		while (modReg.test(modCode)) {
			result[key].isMod = true;
			const [, , quote, name, , c1, , , c2] = modReg.exec(mod.code);
			const code = c1 || c2;
			alter(code, key);
			let nme = name.trim();
			modCode = modCode.replace(modReg, code);
			if (!quote && !nme) continue;
			nme = nme || '/';
			result[key].newName = nme;
		}
		result[key].code = modCode;
	});
	return result;
}
function parseCode(code, options = {}) {
	const result = {};
	groupNumber = 0;
	let currentName = '';
	let internal = '';
	let comment = '';
	let intCnt = 0;
	let cnt = 0;
	let { all } = options;
	let space = '';
	let extra = false;
	let extraStrt = 0;
	if (all != null) all = all || '/';
	const switchModule = name => {
		const nme = !name ? '/' : name;
		if (!result[nme]) {
			result[nme] = {
				name: nme,
				exports: {},
				exportStr: [],
				imports: {},
				importStr: [],
				groups: {},
				internals: {},
				code: '',
				comment,
			};
		}
		cnt = 0;
		currentName = nme;
	};
	switchModule();
	const addChar = char => {
		if (/\s+/.test(char) && comment && char !== '') space += char;
		else {
			if (internal) result[currentName].internals[internal].code += comment + space + char;
			result[currentName].code += comment + space + char;
			comment = '';
			space = '';
		}
	};
	const getDeclare = indx => {
		const declre = helper_1.testModule(code, indx);
		if (declre > 7) {
			if (currentName === '/' && !internal) {
				const name = getName(code, indx + declre + 1);
				if (name.name) {
					currentName = name.name.replace(/['"]/g, '');
					switchModule(currentName);
					if (name.codeType) result[currentName].type = name.codeType;
					return declre + name.full.length;
				}
			} else {
				extra = true;
				extraStrt = internal ? intCnt : cnt;
			}
		}
		if (declre > 0) {
			addChar('declare');
			if (internal) result[currentName].internals[internal].hasDeclare = true;
			else result[currentName].hasDeclare = true;
			return 6;
		}
		addChar('d');
		return 0;
	};
	const addMem = props => {
		const p = { ...props };
		const grp = props.complex && true;
		if (grp) p.group = true;
		if (p.name === '*') p.all = true;
		if (comment) {
			if (props.len === 1) p.comment = comment + space;
			else result[currentName].code += comment + space;
		}
		comment = '';
		space = '';
		const name = props.as || props.name;
		if (props.portType === 'im') {
			if (name && name !== '*') result[currentName].imports[name] = p;
			else result[currentName].importStr.push(p.from);
		} else if (name !== '*') result[currentName].exports[name] = p;
		else result[currentName].exportStr.push(p.from);
		if (grp) {
			if (!result[currentName].groups[p.positionID])
				result[currentName].groups[p.positionID] = [name];
			else {
				result[currentName].groups[p.positionID].push(name);
				return;
			}
		}
		result[currentName].code += `\n--replace${p.positionID}=`;
		if (grp) result[currentName].code += `${p.from}[grp`;
		else if (name && name !== '*') result[currentName].code += `${name}[`;
		else if ((name === '*' || (!name && props.portType === 'im')) && props.from) {
			result[currentName].code += `${p.from}[all`;
		}
		result[currentName].code += `${props.portType}]`;
	};
	const portTest = indx => {
		const portProp = helper_1.getPort(code, indx);
		if (portProp) {
			if (portProp === true) {
				const name = getName(code, indx + 6);
				if (name.name) {
					result[currentName].internals[name.name] = {
						owner: currentName,
						name: name.name,
						type: name.codeType,
						comment,
						prename: name.pre,
						portType: 'ex',
						code: name.codeType === types_1.codeType.roundBracket ? '(' : '',
					};
					internal = name.name;
					if (space) {
						result[currentName].internals[internal].comment += space;
						result[currentName].code += comment + space;
						comment = '';
						space = '';
					}
					result[currentName].code += `${comment}export${name.full}`;
					comment = '';
					return 6 + name.full.length - 1;
				}
				return 0;
			}
			const prop = { ...portProp.member, positionID: getGroupID(currentName) };
			if (portProp.member.complex) {
				const cplxSplit = portProp.member.complex.split(',');
				const cplxLen = cplxSplit.length;
				cplxSplit.forEach(c =>
					addMem({ ...prop, ...helper_1.getNewName(c), len: cplxLen }),
				);
			} else if (portProp.member.name) addMem({ ...prop, len: 1 });
			else addMem({ ...prop, len: 0 });
			return Math.max(portProp.str.length - 1, 0);
		}
		return 0;
	};
	for (let i = 0, len = code.length; i < len; i++) {
		const char = code[i];
		switch (char) {
			case '(':
			case '{':
				if (internal) intCnt++;
				else cnt++;
				addChar(char);
				break;
			case ')':
				if (internal) intCnt--;
				else cnt--;
				addChar(char);
				break;
			case '}':
				if (internal) intCnt--;
				else cnt--;
			case ';':
				if (extra) {
					if (internal && extraStrt === intCnt) extra = false;
					else if (!internal && extraStrt === cnt) extra = false;
				}
				if (internal) {
					if (
						intCnt < 0 ||
						result[currentName].internals[internal].type == null ||
						(result[currentName].internals[internal].type !==
							types_1.codeType.curlyBracket &&
							intCnt === 0)
					) {
						if (
							result[currentName].internals[internal].type !==
							types_1.codeType.curlyBracket
						)
							result[currentName].internals[internal].code += char;
						internal = '';
						intCnt = 0;
					}
					addChar(char);
				} else if (cnt < 0 || result[currentName].type === types_1.codeType.equal) {
					switchModule();
				} else addChar(char);
				break;
			case '/':
				if (comment) addChar('');
				const c = helper_1.getComment(code, i);
				if (c.length === 1) addChar('/');
				if (c.length > 1) {
					i += c.length - 1;
					if (/\/\*\*\s*\$mod(end|=|\s*\*\/)/.test(c)) addChar(c);
					else comment = c;
				}
				break;
			case 'd':
				if (extra) addChar(char);
				else i += getDeclare(i);
				comment = '';
				break;
			case 'i':
			case 'e':
				if (all || internal || extra) addChar(char);
				else {
					const test = portTest(i);
					if (test === 0) addChar(char);
					else i += test;
				}
				break;
			default:
				addChar(char);
				break;
		}
	}
	return rename(getExportValues(result), options);
}
exports.parseCode = parseCode;
